/* This is the template file for HW1_7. In this exercise, you should 
 * implement a code to compute
 *         b = A*B*x
 * in two different ways:
 * 1) Implement it as b = A*(B*x).
 * 2) Implement it as b = (A*B)*x.
 * Compare the efficiency of the two different approaches.
 * 
 * Note: This template is incomplete. You need to complete it by implementing
 *   the part marked by FIXME. 
 *
 * Submit your completed code to the TA by email. Also submit a report for an 
 * analysis of the numbers of floating point operations required by the two 
 * approaches. Do your analysis correlates well with their actual performance?
 *
 * To compile, use UNIX command 
 *      cc -O -Wall -o hw1_7 hw1_7.c -lm
 * This will generate an executable hw1_7. The -O option is optional, and it
 * invokes compiler optimization. If you want to use a debugger, change -O by -g. 
 * The -Wall option would enable compiler's warning messages.
 *
 * Run the program in the smalltest or largetest subdirectory with UNIX commands
 *      cd smalltest
 *      ../hw1_7
 *      cd ../largetest
 *      ../hw1_7
 * which would print out the answers (or part of the answer) to the screen.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/* Define types for matrix and vector. 
 * Access the (i,j)-entries of a matrix A using A.vals[i][j], and 
 * access the ith entries of a vector x using x.vals[i], where 
 * i and j start from 0.
 */

typedef struct {
    double** vals;
    int m; /* m is the number of rows */
    int n; /* n is the number of columns */
} Matrix;

typedef struct {
    double*  vals;
    int m;
} Vector;


/* Define prototypes of helper functions */
Matrix allocate_matrix(int m, int n);
Vector allocate_vector( int m);
double tictoc (int n);


/*-----------------------------------------------------------------------------
 * FUNCTION: matrix_vector_mul - matrix-vector multiplication A*x
 * --------------------------------------------------------------------------*/
Vector matrix_vector_mul( Matrix A, Vector x) {
    /* Allocate memory for vector b */
    Vector b = allocate_vector( A.m);
    int i, j;

    /* FIXME: Compute matrix-vector multiplication b=A*x */
    for (i=0; i<A.m; i++) {
        for (j=0; j<A.n; j++) {

        }
    }

    return b;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: matrix_matrix_mul - matrix-matrix multiplication A*B
 * --------------------------------------------------------------------------*/
Matrix matrix_matrix_mul( Matrix A, Matrix B) {
    /* Allocate memory for C */
    Matrix C = allocate_matrix( A.m, B.n);
    int i,j,k;

    /* FIXME: Compute matrix-matrix multiplication C=A*B */

    return C;
}

/* Define prototypes of additional helper functions */
void   deallocate_matrix( Matrix);
void   deallocate_vector( Vector);

Matrix read_matrix(const char *);
Vector read_vector(const char *);
void   print_vector( Vector);

/*-----------------------------------------------------------------------------
 * FUNCTION: main - main driver routine.
 * --------------------------------------------------------------------------*/
int main( int argc, char **argv) {
    Matrix A, B, C;
    Vector x, y, b;
    int i, N=1000;

    /* Read in matrices and vectors */
    A = read_matrix( "A.txt");
    B = read_matrix( "B.txt");
    x = read_vector( "x.txt");

    /* Method 1: compute y = B*x, and then b = A * y. 
     * For timing purpose, run for N times and then divide time by N.
     */
    tictoc(1);
    for (i=0; i<N; i++) {
        y = matrix_vector_mul( B, x);
        b = matrix_vector_mul( A, y); 

        deallocate_vector( y);
        if (i!=N-1) deallocate_vector( b);
    }
    printf( "Method 1 took %g seconds.\n", tictoc(2)/N);
    
    print_vector( b);
    deallocate_vector( b);

    /* Method 2: compute C = A*B, and then b = C * x
     */
    tictoc(1);
    for (i=0; i<N; i++) {
        C = matrix_matrix_mul( A, B);
        b = matrix_vector_mul( C, x); 

        deallocate_matrix( C);
        if (i!=N-1) deallocate_vector( b);
    }
    printf( "Method 2 took %g seconds.\n", tictoc(2)/N);

    print_vector( b);
    deallocate_vector( b);

    /* De-allocate the matrix and vectors */
    deallocate_matrix( A);
    deallocate_matrix( B);
    deallocate_vector( x);

    return 0;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: allocate_matrix - Allocate memory for a given matrix.
 * --------------------------------------------------------------------------*/
/* Disclaimer: The approach used here is not the most efficient way for
 * implementing matrices, but it is adopted here for convenience.
 */
Matrix allocate_matrix(int m, int n) {
    Matrix A;
    double *ptmp;
    int i;

    A.m = m; A.n = n;
    ptmp = (double *)malloc( sizeof(double) * m * n);
    A.vals = (double **)malloc( sizeof(double **) * m);

    for (i=0; i<m; ++i) {
        A.vals[i] = ptmp + i*n;
    }

    return A;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: deallocate_matrix - De-allocate memory for a matrix.
 * --------------------------------------------------------------------------*/
void deallocate_matrix( Matrix A) {
    free( A.vals[0]);
    free( A.vals); A.vals=NULL;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: read_matrix - Read in a matrix from file.
 * --------------------------------------------------------------------------*/
Matrix read_matrix(const char *fname) {
    FILE *fin;
    int m, n, i, j;
    Matrix A;

    fin = fopen( fname, "r");
    if (!fin) {
        printf("Could not open file %s. Aborting...\n", fname);
        exit(-1);
    }
    fscanf( fin, "%d %d", &m, &n);
    A = allocate_matrix( m, n);

    for (i=0; i<m; ++i) {
        for (j=0; j<n; ++j) {
            fscanf( fin, "%lf", &A.vals[i][j]);
        }
    }

    fclose(fin);
    return A;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: read_vector - Read in a vector from file.
 * --------------------------------------------------------------------------*/
Vector read_vector(const char *fname) {
    FILE *fin;
    int m, i;
    Vector x;

    fin = fopen( fname, "r");
    if (!fin) {
        printf("Could not open file %s. Aborting...\n", fname);
        exit(-1);
    }
    fscanf( fin, "%d", &m);
    x = allocate_vector( m);

    for (i=0; i<m; ++i) {
        fscanf( fin, "%lf", &x.vals[i]);
    }

    fclose(fin);
    return x;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: allocate_vector - Allocate memory for an m vector.
 * --------------------------------------------------------------------------*/
Vector allocate_vector( int m) {
    Vector x;

    x.m = m;
    x.vals = (double *)malloc( sizeof(double) * m);
    return x;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: deallocate_vector - De-allocate the memory for an m vector. 
 * --------------------------------------------------------------------------*/
void deallocate_vector( Vector x) {
    free( x.vals); x.vals=NULL;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: print_vector - Print a vector. Only print the first 20 entries.
 * --------------------------------------------------------------------------*/
void print_vector( Vector x) {
    int i, m;

    m = (x.m<20) ? x.m : 20; 

    printf("ans = \n");
    for (i=0; i<m; ++i) {
        printf("      %lf\n", x.vals[i]);
    }
}

#include <time.h>
#include <sys/time.h>

/*-----------------------------------------------------------------------------
 * FUNCTION: tictoc - Initialize (1) or get (2) the elapsed time in seconds.
 * --------------------------------------------------------------------------*/
double tictoc (int n)
{
    double    y = 0.0;
    static    struct timeval start_time, end_time; 

    if (n == 1) { 
	gettimeofday(&start_time, NULL);
    }
    else if (n == 2) { 
	gettimeofday(&end_time, NULL);
        
	y = (end_time.tv_sec - start_time.tv_sec) + 
            (end_time.tv_usec-start_time.tv_usec)*1.e-6;
    }
    return (y); 
}
